from sqlalchemy import func
from tlbx import st

from zillion.dialects.conversions import DialectDateConversions


class SQLiteDialectDateConversions(DialectDateConversions):
    @classmethod
    def date_year_start(cls, x):
        return func.DATE(str(x) + "-01-01")

    @classmethod
    def date_year_plus_year(cls, x):
        return func.DATE(str(x) + "-01-01", "+1 year")

    @classmethod
    def datetime_year_end(cls, x):
        return func.DATETIME(str(x) + "-01-01", "+1 year", "-1 second")

    @classmethod
    def date_month_start(cls, x):
        return func.DATE(str(x) + "-01")

    @classmethod
    def date_month_plus_month(cls, x):
        return func.DATE(str(x) + "-01", "+1 month")

    @classmethod
    def datetime_month_end(cls, x):
        return func.DATETIME(str(x) + "-01", "+1 month", "-1 second")

    @classmethod
    def date_plus_day(cls, x):
        return func.DATE(x, "+1 day")

    @classmethod
    def datetime_day_end(cls, x):
        return func.DATETIME(x, "+1 day", "-1 second")

    @classmethod
    def datetime_hour_plus_hour(cls, x):
        return func.DATETIME(x, "+1 hour")

    @classmethod
    def datetime_hour_end(cls, x):
        return func.DATETIME(x, "+1 hour", "-1 second")

    @classmethod
    def datetime_minute_plus_minute(cls, x):
        return func.DATETIME(x, "+1 minute")

    @classmethod
    def datetime_minute_end(cls, x):
        return func.DATETIME(x, "+1 minute", "-1 second")


SQLITE_DIALECT_CONVERSIONS = {
    "year": {
        "ds_formula": "cast(strftime('%Y', {}) as integer)",
        "ds_criteria_conversions": SQLiteDialectDateConversions.get_year_criteria_conversions(),
    },
    "quarter": "strftime('%Y', {}) || '-Q' || ((cast(strftime('%m', {}) as integer) + 2) / 3)",  # 2020-Q1
    "quarter_of_year": "(cast(strftime('%m', {}) as integer) + 2) / 3",
    "month": {
        "ds_formula": "strftime('%Y-%m', {})",
        "ds_criteria_conversions": SQLiteDialectDateConversions.get_month_criteria_conversions(),
    },
    "month_name": (
        "CASE strftime('%m', {}) "
        "WHEN '01' THEN 'January' "
        "WHEN '02' THEN 'February' "
        "WHEN '03' THEN 'March' "
        "WHEN '04' THEN 'April' "
        "WHEN '05' THEN 'May' "
        "WHEN '06' THEN 'June' "
        "WHEN '07' THEN 'July' "
        "WHEN '08' THEN 'August' "
        "WHEN '09' THEN 'September' "
        "WHEN '10' THEN 'October' "
        "WHEN '11' THEN 'November' "
        "WHEN '12' THEN 'December' "
        "ELSE NULL "
        "END"
    ),
    "month_of_year": "cast(strftime('%m', {}) as integer)",
    "week_of_year": "cast(strftime('%W', {}) as integer)",
    "date": {
        "ds_formula": "strftime('%Y-%m-%d', {})",
        "ds_criteria_conversions": SQLiteDialectDateConversions.get_date_criteria_conversions(),
    },
    "day_name": (
        "CASE cast(strftime('%w', {}) as integer) "
        "WHEN 0 THEN 'Sunday' "
        "WHEN 1 THEN 'Monday' "
        "WHEN 2 THEN 'Tuesday' "
        "WHEN 3 THEN 'Wednesday' "
        "WHEN 4 THEN 'Thursday' "
        "WHEN 5 THEN 'Friday' "
        "WHEN 6 THEN 'Saturday' "
        "ELSE NULL "
        "END"
    ),
    "day_of_week": "(cast(strftime('%w', {}) as integer) + 6) % 7 + 1",  # Convert to Monday = 1
    "is_weekday": (
        "CASE cast(strftime('%w', {}) as integer) "
        "WHEN 0 THEN 0 "
        "WHEN 1 THEN 1 "
        "WHEN 2 THEN 1 "
        "WHEN 3 THEN 1 "
        "WHEN 4 THEN 1 "
        "WHEN 5 THEN 1 "
        "WHEN 6 THEN 0 "
        "ELSE NULL "
        "END"
    ),
    "day_of_month": "cast(strftime('%d', {}) as integer)",
    "day_of_year": "cast(strftime('%j', {}) as integer)",
    "hour": {
        "ds_formula": "strftime('%Y-%m-%d %H:00:00', {})",
        "ds_criteria_conversions": SQLiteDialectDateConversions.get_hour_criteria_conversions(),
    },
    "hour_of_day": "cast(strftime('%H', {}) as integer)",
    "minute": {
        "ds_formula": "strftime('%Y-%m-%d %H:%M:00', {})",
        "ds_criteria_conversions": SQLiteDialectDateConversions.get_minute_criteria_conversions(),
    },
    "minute_of_hour": "cast(strftime('%M', {}) as integer)",
    "datetime": {
        "ds_formula": "strftime('%Y-%m-%d %H:%M:%S', {})",
        "ds_criteria_conversions": SQLiteDialectDateConversions.get_datetime_criteria_conversions(),
    },
    "unixtime": "cast(strftime('%s', {}) as integer)",
}



from sqlalchemy import func, text

from zillion.dialects.conversions import DialectDateConversions


class MySQLDialectDateConversions(DialectDateConversions):
    @classmethod
    def date_year_start(cls, x):
        return func.CONCAT(x, "-01-01")

    @classmethod
    def date_year_plus_year(cls, x):
        return func.DATE_ADD(str(x) + "-01-01", text("INTERVAL 1 YEAR"))

    @classmethod
    def datetime_year_end(cls, x):
        return func.DATE_SUB(
            func.DATE_ADD(str(x) + "-01-01", text("INTERVAL 1 YEAR")),
            text("INTERVAL 1 SECOND"),
        )

    @classmethod
    def date_month_start(cls, x):
        return func.CONCAT(x, "-01")

    @classmethod
    def date_month_plus_month(cls, x):
        return func.DATE_ADD(str(x) + "-01", text("INTERVAL 1 MONTH"))

    @classmethod
    def datetime_month_end(cls, x):
        return func.DATE_SUB(
            func.DATE_ADD(str(x) + "-01", text("INTERVAL 1 MONTH")),
            text("INTERVAL 1 SECOND"),
        )

    @classmethod
    def date_plus_day(cls, x):
        return func.DATE_ADD(x, text("INTERVAL 1 DAY"))

    @classmethod
    def datetime_day_end(cls, x):
        return func.DATE_SUB(
            func.DATE_ADD(x, text("INTERVAL 1 DAY")), text("INTERVAL 1 SECOND")
        )

    @classmethod
    def datetime_hour_plus_hour(cls, x):
        return func.DATE_ADD(x, text("INTERVAL 1 HOUR"))

    @classmethod
    def datetime_hour_end(cls, x):
        return func.DATE_SUB(
            func.DATE_ADD(x, text("INTERVAL 1 HOUR")), text("INTERVAL 1 SECOND")
        )

    @classmethod
    def datetime_minute_plus_minute(cls, x):
        return func.DATE_ADD(x, text("INTERVAL 1 MINUTE"))

    @classmethod
    def datetime_minute_end(cls, x):
        return func.DATE_SUB(
            func.DATE_ADD(x, text("INTERVAL 1 MINUTE")), text("INTERVAL 1 SECOND")
        )


MYSQL_DIALECT_CONVERSIONS = {
    "year": {
        "ds_formula": "EXTRACT(YEAR FROM {})",
        "ds_criteria_conversions": MySQLDialectDateConversions.get_year_criteria_conversions(),
    },
    "quarter": "CONCAT(YEAR({}), '-Q', QUARTER({}))",
    "quarter_of_year": "EXTRACT(QUARTER FROM {})",
    "month": {
        "ds_formula": "DATE_FORMAT({}, '%Y-%m')",
        "ds_criteria_conversions": MySQLDialectDateConversions.get_month_criteria_conversions(),
    },
    "month_name": "MONTHNAME({})",
    "month_of_year": "EXTRACT(MONTH FROM {})",
    "week_of_year": "EXTRACT(WEEK FROM {})",
    "date": {
        "ds_formula": "DATE_FORMAT({}, '%Y-%m-%d')",
        "ds_criteria_conversions": MySQLDialectDateConversions.get_date_criteria_conversions(),
    },
    "day_name": "DAYNAME({})",
    "day_of_week": "WEEKDAY({}) + 1",  # Monday = 1
    "is_weekday": "IF((WEEKDAY({}) + 1) < 6, 1, 0)",  # Monday = 1
    "day_of_month": "EXTRACT(DAY FROM {})",
    "day_of_year": "DAYOFYEAR({})",
    "hour": {
        "ds_formula": "DATE_FORMAT({}, '%Y-%m-%d %H:00:00')",
        "ds_criteria_conversions": MySQLDialectDateConversions.get_hour_criteria_conversions(),
    },
    "hour_of_day": "EXTRACT(HOUR FROM {})",
    "minute": {
        "ds_formula": "DATE_FORMAT({}, '%Y-%m-%d %H:%i:00')",
        "ds_criteria_conversions": MySQLDialectDateConversions.get_minute_criteria_conversions(),
    },
    "minute_of_hour": "EXTRACT(MINUTE FROM {})",
    "datetime": {
        "ds_formula": "DATE_FORMAT({}, '%Y-%m-%d %H:%i:%S')",
        "ds_criteria_conversions": MySQLDialectDateConversions.get_datetime_criteria_conversions(),
    },
    "unixtime": "UNIX_TIMESTAMP({})",
}



import sqlalchemy as sa

from zillion.core import zillion_config

zillion_engine = sa.create_engine(zillion_config["DB_URL"], pool_pre_ping=True)
zillion_metadata = sa.MetaData()
zillion_metadata.bind = zillion_engine

Warehouses = sa.Table(
    "warehouses",
    zillion_metadata,
    sa.Column("id", sa.Integer, primary_key=True),
    sa.Column("name", sa.String(128), nullable=False, unique=True),
    sa.Column("params", sa.Text, nullable=False),
    sa.Column("meta", sa.Text),
    sa.Column("created_at", sa.DateTime, server_default=sa.func.NOW()),
)

ReportSpecs = sa.Table(
    "report_specs",
    zillion_metadata,
    sa.Column("id", sa.Integer, primary_key=True),
    sa.Column("warehouse_id", sa.Integer, nullable=False),
    sa.Column("params", sa.Text, nullable=False),
    sa.Column("meta", sa.Text),
    sa.Column("created_at", sa.DateTime, server_default=sa.func.NOW()),
)

DimensionValues = sa.Table(
    "dimension_values",
    zillion_metadata,
    sa.Column("name", sa.String(100), primary_key=True),
    sa.Column("warehouse_id", sa.Integer, primary_key=True),
    sa.Column("values", sa.Text, nullable=False),
    sa.Column("created_at", sa.DateTime, server_default=sa.func.NOW()),
)

zillion_metadata.create_all(zillion_engine)



import sqlalchemy as sa


class DialectDateConversions:
    @classmethod
    def f(cls, func, i=0):
        """Generate a callable of a conversion function taking a specific index from
        the input data"""
        return lambda x, cls=cls, i=i: getattr(cls, func)(x[i])

    @classmethod
    def raw_value(cls, v):
        return sa.literal(v)

    @classmethod
    def date_year_start(cls, v):
        raise NotImplementedError

    @classmethod
    def date_year_plus_year(cls, v):
        raise NotImplementedError

    @classmethod
    def datetime_year_end(cls, x):
        raise NotImplementedError

    @classmethod
    def date_month_start(cls, x):
        raise NotImplementedError

    @classmethod
    def date_month_plus_month(cls, x):
        raise NotImplementedError

    @classmethod
    def datetime_month_end(cls, x):
        raise NotImplementedError

    @classmethod
    def date_plus_day(cls, x):
        raise NotImplementedError

    @classmethod
    def datetime_day_end(cls, x):
        raise NotImplementedError

    @classmethod
    def datetime_hour_plus_hour(cls, x):
        raise NotImplementedError

    @classmethod
    def datetime_hour_end(cls, x):
        raise NotImplementedError

    @classmethod
    def datetime_minute_plus_minute(cls, x):
        raise NotImplementedError

    @classmethod
    def datetime_minute_end(cls, x):
        raise NotImplementedError

    @classmethod
    def get_year_criteria_conversions(cls):
        return {
            "=": [
                [">=", cls.f("date_year_start")],
                ["<", cls.f("date_year_plus_year")],
            ],
            # NOTE: all of the conditions in the criteria replacement lists get AND'd together
            # so we can't simply do "x < 2020-01-01 or x >= 2021-01-01"
            "!=": [
                ["not between", [cls.f("date_year_start"), cls.f("datetime_year_end")]]
            ],
            ">": [[">=", cls.f("date_year_plus_year")]],
            ">=": [[">=", cls.f("date_year_start")]],
            "<": [["<", cls.f("date_year_start")]],
            "<=": [["<", cls.f("date_year_plus_year")]],
            "between": [
                [">=", cls.f("date_year_start")],
                ["<", cls.f("date_year_plus_year", i=1)],
            ],
            "not between": [
                [
                    "not between",
                    [cls.f("date_year_start"), cls.f("datetime_year_end", i=1)],
                ]
            ],
        }

    @classmethod
    def get_month_criteria_conversions(cls):
        return {
            "=": [
                [">=", cls.f("date_month_start")],
                ["<", cls.f("date_month_plus_month")],
            ],
            "!=": [
                [
                    "not between",
                    [cls.f("date_month_start"), cls.f("datetime_month_end")],
                ]
            ],
            ">": [[">=", cls.f("date_month_plus_month")]],
            ">=": [[">=", cls.f("date_month_start")]],
            "<": [["<", cls.f("date_month_start")]],
            "<=": [["<", cls.f("date_month_plus_month")]],
            "between": [
                [">=", cls.f("date_month_start")],
                ["<", cls.f("date_month_plus_month", i=1)],
            ],
            "not between": [
                [
                    "not between",
                    [cls.f("date_month_start"), cls.f("datetime_month_end", i=1)],
                ]
            ],
        }

    @classmethod
    def get_date_criteria_conversions(cls):
        return {
            "=": [[">=", cls.f("raw_value")], ["<", cls.f("date_plus_day")]],
            "!=": [["not between", [cls.f("raw_value"), cls.f("datetime_day_end")]]],
            ">": [[">=", cls.f("date_plus_day")]],
            ">=": [[">=", cls.f("raw_value")]],
            "<": [["<", cls.f("raw_value")]],
            "<=": [["<", cls.f("date_plus_day")]],
            "between": [[">=", cls.f("raw_value")], ["<", cls.f("date_plus_day", i=1)]],
            "not between": [
                ["not between", [cls.f("raw_value"), cls.f("datetime_day_end", i=1)]]
            ],
        }
        # return {
        #     "=": [[">=", ":0"], ["<", cls.f("date_plus_day")]],
        #     "!=": [["not between", [":0", cls.f("datetime_day_end")]]],
        #     ">": [[">=", cls.f("date_plus_day")]],
        #     ">=": [[">=", ":0"]],
        #     "<": [["<", ":0"]],
        #     "<=": [["<", cls.f("date_plus_day")]],
        #     "between": [[">=", ":0"], ["<", cls.f("date_plus_day", i=1)]],
        #     "not between": [["not between", [":0", cls.f("datetime_day_end", i=1)]]],
        # }

    @classmethod
    def get_hour_criteria_conversions(cls):
        return {
            "=": [[">=", cls.f("raw_value")], ["<", cls.f("datetime_hour_plus_hour")]],
            "!=": [["not between", [cls.f("raw_value"), cls.f("datetime_hour_end")]]],
            ">": [[">=", cls.f("datetime_hour_plus_hour")]],
            ">=": [[">=", cls.f("raw_value")]],
            "<": [["<", cls.f("raw_value")]],
            "<=": [["<", cls.f("datetime_hour_plus_hour")]],
            "between": [
                [">=", cls.f("raw_value")],
                ["<", cls.f("datetime_hour_plus_hour", i=1)],
            ],
            "not between": [
                ["not between", [cls.f("raw_value"), cls.f("datetime_hour_end", i=1)]]
            ],
        }

    @classmethod
    def get_minute_criteria_conversions(cls):
        return {
            "=": [
                [">=", cls.f("raw_value")],
                ["<", cls.f("datetime_minute_plus_minute")],
            ],
            "!=": [["not between", [cls.f("raw_value"), cls.f("datetime_minute_end")]]],
            ">": [[">=", cls.f("datetime_minute_plus_minute")]],
            ">=": [[">=", cls.f("raw_value")]],
            "<": [["<", cls.f("raw_value")]],
            "<=": [["<", cls.f("datetime_minute_plus_minute")]],
            "between": [
                [">=", cls.f("raw_value")],
                ["<", cls.f("datetime_minute_plus_minute", i=1)],
            ],
            "not between": [
                ["not between", [cls.f("raw_value"), cls.f("datetime_minute_end", i=1)]]
            ],
        }

    @classmethod
    def get_datetime_criteria_conversions(cls):
        return {
            "=": [["=", cls.f("raw_value")]],
            "!=": [["!=", cls.f("raw_value")]],
            ">": [[">", cls.f("raw_value")]],
            ">=": [[">=", cls.f("raw_value")]],
            "<": [["<", cls.f("raw_value")]],
            "<=": [["<=", cls.f("raw_value")]],
            "between": [["between", [cls.f("raw_value"), cls.f("raw_value", i=1)]]],
            "not between": [
                ["not between", [cls.f("raw_value"), cls.f("raw_value", i=1)]]
            ],
        }
